use crate::read;
use itertools::Itertools;

#[derive(Debug, PartialEq, Eq, Clone, Hash, Default)]
struct Coord {
    pub row: usize,
    pub col: usize,
}

pub fn run() {
    let input = read("12");
    let lines = input.lines();
    let mut start: Coord = Default::default();
    let mut target: Coord = Default::default();
    let map = lines
        .enumerate()
        .map(|(r, l)| {
            l.chars()
                .enumerate()
                .map(|(col, c)| match c {
                    'S' => {
                        start = Coord {
                            row: r,
                            col,
                        };
                        0
                    }
                    'E' => {
                        target = Coord {
                            row: r,
                            col,
                        };
                        25
                    }
                    c => (c as u8) - ('a' as u8),
                })
                .collect_vec()
        })
        .collect_vec();

    #[cfg(feature = "part1")]
    {
        println!("Day 12, Part 01: {:?}", bfs(&map, start, target.clone()).unwrap());
    }

    #[cfg(feature = "part2")]
    {
        let starting_points = map
            .iter()
            .enumerate()
            .flat_map(|(row, v)| {
                v.iter()
                    .enumerate()
                    .filter(|(_, vv)| **vv == 0)
                    .map(|(col, _)| Coord {
                        row,
                        col,
                    })
                    .collect_vec()
            })
            .collect_vec();
        let min = starting_points
            .iter()
            .filter_map(|start| bfs(&map, start.clone(), target.clone()))
            .min()
            .unwrap();
        println!("Day 12, Part 02: {:?}", min);
    }
}

fn bfs(map: &Vec<Vec<u8>>, start: Coord, target: Coord) -> Option<usize> {
    let mut visited = vec![vec![false; map[0].len()]; map.len()];
    let mut queue = std::collections::VecDeque::new();
    queue.push_back((start, 0));

    while let Some((next, path_length)) = queue.pop_front() {
        if next == target {
            return Some(path_length);
        }
        match make_coord((next.row as isize) + 1, next.col as isize, &map, &visited) {
            Some(down) if map[next.row][next.col] + 1
                >= map[down.row][down.col] => {
                visited[down.row][down.col] = true;
                queue.push_back((down, path_length + 1));
            },
            _ => {}
        }
        match make_coord((next.row as isize) - 1, next.col as isize, &map, &visited) {
            Some(up) if map[next.row][next.col] + 1
                >= map[up.row][up.col] => {
                visited[up.row][up.col] = true;
                queue.push_back((up, path_length + 1));
            },
            _ => {}
        }
        match make_coord(next.row as isize, (next.col as isize) - 1, &map, &visited) {
            Some(left) if map[next.row][next.col] + 1
                >= map[left.row][left.col] => {
                visited[left.row][left.col] = true;
                queue.push_back((left, path_length + 1));
            },
            _ => {}
        }
        match make_coord(next.row as isize, (next.col as isize) + 1, &map, &visited) {
            Some(right) if map[next.row][next.col] + 1
                >= map[right.row][right.col] => {
                visited[right.row][right.col] = true;
                queue.push_back((right, path_length + 1));
            },
            _ => {}
        }
    }
    None
}

fn make_coord(row: isize, col: isize, map: &Vec<Vec<u8>>, visited: &Vec<Vec<bool>>) -> Option<Coord> {
    if row >= 0
        && col >= 0
        && row < map.len() as isize
        && (col as usize) < map[row as usize].len()
        && !visited[row as usize][col as usize] {
        Some(Coord {
            row: row as usize,
            col: col as usize
        })
    } else {
        None
    }
}
